@@ -402,3 +402,13 @@ test_that("subset argument", {
                  c(`(Intercept)` = 248.63573143019875,
                    Days = 10.904528975303338))
 })
+
+test_that("start argument is a named list", {
+  expect_error(glmmTMB(mpg ~ hp, data = mtcars, start = c(0, 1)),
+               "'start' should be a named list")
+})
+
+test_that("start argument has correct elements", {
+  expect_error(glmmTMB(mpg ~ hp, data = mtcars, start = list(junk = 1)),
+               "unrecognized vector")
+})
